package algorithms;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import methods.Method;
import objects.Classifiable;
import objects.ClassifiableList;
import objects.DistanceMatrix;
import objects.Motion;
import objects.Template;

public abstract class MethodEvaluator {

    private final Method[] methods;
    private final List<Classifiable> classifiables;
    private final File directoryAMC;

    private Classifier classifier;
    private DistanceMatrix distanceMatrix;

    public MethodEvaluator(Method[] methods, File directoryAMC) {
        this.methods = methods;
        this.directoryAMC = directoryAMC;
        classifiables = new ArrayList();
        for (File fileAMC : directoryAMC.listFiles()) {
            classifiables.add(new Classifiable(fileAMC.getName().split("\\.")[0]));
        }
    }

    public List<Classifiable> getClassifiables() {
        return classifiables;
    }

    public DistanceMatrix getDistanceMatrix() {
        return distanceMatrix;
    }

    public Classifier getClassifier() {
        return classifier;
    }

    public void setClassifier(Classifier classifier) {
        this.classifier = classifier;
    }

    public int getNumberOfMethods() {
        return methods.length;
    }

    public Method getMethod(int m) {
        return methods[m];
    }

    public abstract void initialize();

    public abstract void evaluate() throws IOException;

    public abstract void print();

    public List<Motion> loadMotions(Method method, List<String> subjects) throws IOException {
        List<Motion> motions = new ArrayList();
        for (File fileAMC : directoryAMC.listFiles()) {
            if (subjects.contains(fileAMC.getName().split("_")[0])) {
                motions.add(method.loadMotion(fileAMC));
            }
        }
        return motions;
    }

    public DistanceMatrix calculateDistanceMatrix(List<Template> templates) {
        distanceMatrix = new DistanceMatrix(classifier.getDecision().getDistanceTemplates());
        for (Template template1 : templates) {
            for (Template template2 : templates) {
                distanceMatrix.putDistance(template1, template2);
            }
        }
        return distanceMatrix;
    }

    public String toString(List<Double> scores) {
        return (double) Math.round(getAverage(scores) * 10000) / 10000 + ",";
    }

    public List<double[]> sortCurve(List<double[]> curve, int index) {
        List<double[]> sortedCurve = new ArrayList();
        int length = curve.size();
        for (int l = 0; l < length; l++) {
            double minPointValue = Double.MAX_VALUE;
            int minPointIndex = 0;
            for (int pointIndex = 0; pointIndex < curve.size(); pointIndex++) {
                if (minPointValue > curve.get(pointIndex)[index]) {
                    minPointValue = curve.get(pointIndex)[index];
                    minPointIndex = pointIndex;
                }
            }
            sortedCurve.add(curve.get(minPointIndex));
            curve.remove(minPointIndex);
        }
        return sortedCurve;
    }

    public double getAverage(List<Double> scores) {
        double sum = 0;
        for (double score : scores) {
            sum += score;
        }
        return sum / scores.size();
    }

    public double[] getCumulativeMatchCharacteristic(List<Template> templatesTest, List<Template> templatesGallery) {
        List<List<Template>> templatesBySubject = new ClassifiableList(templatesGallery).splitBySubject();
        int numberOfSubjects = templatesBySubject.size();
        int number = templatesTest.size();
        double[] cmc = new double[numberOfSubjects];
        for (Template templateTest : templatesTest) {
            List<Template> templatesGalleryRank = new ArrayList();
            templatesGalleryRank.addAll(templatesGallery);
            String[] ranking = new String[numberOfSubjects];
            for (int rank = 0; rank < numberOfSubjects; rank++) {
                classifier.getDecision().importGallery(templatesGalleryRank);
                String classifiedSubject = classifier.classify(templateTest);
                ranking[rank] = classifiedSubject;
                int i = 0;
                while (!templatesBySubject.get(i).get(0).getSubject().equals(classifiedSubject)) {
                    i++;
                }
                templatesGalleryRank.removeAll(templatesBySubject.get(i));
            }
            for (int rank = 0; rank < numberOfSubjects; rank++) {
                boolean isIn = false;
                for (int r = 0; r <= rank; r++) {
                    if (ranking[r].equals(templateTest.getSubject())) {
                        isIn = true;
                    }
                }
                if (isIn) {
                    cmc[rank] += (double) 1 / number;
                }
            }
        }
        return cmc;
    }

    public double getDaviesBouldinIndex(List<Template> templatesEvaluation) { // low is good
        List<List<Template>> templatesBySubject = new ClassifiableList(templatesEvaluation).splitBySubject();
        int numberOfSubjects = templatesBySubject.size();
        double sum = 0;
        for (int c1 = 0; c1 < numberOfSubjects; c1++) {
            List<Template> templatesOfSubject1 = templatesBySubject.get(c1);
            Template centroid1 = getCentroid(templatesOfSubject1);
            double avgDistanceToCentroid1 = getAverageDistanceToTemplate(templatesOfSubject1, centroid1);
            double maxFrac = 0;
            for (int c2 = c1 + 1; c2 < numberOfSubjects; c2++) {
                List<Template> templatesOfSubject2 = templatesBySubject.get(c2);
                Template centroid2 = getCentroid(templatesOfSubject2);
                double avgDistanceToCentroid2 = getAverageDistanceToTemplate(templatesOfSubject2, centroid2);
                double frac = (avgDistanceToCentroid1 + avgDistanceToCentroid2) / getDistance(centroid1, centroid2);
                if (maxFrac < frac) {
                    maxFrac = frac;
                }
            }
            sum += maxFrac;
        }
        return sum / numberOfSubjects;
    }

    public double getDunnIndex(List<Template> templatesEvaluation) { // high is good
        List<List<Template>> templatesBySubject = new ClassifiableList(templatesEvaluation).splitBySubject();
        int numberOfSubjects = templatesBySubject.size();
        double min = Double.MAX_VALUE;
        for (int c1 = 0; c1 < numberOfSubjects; c1++) {
            Template centroid1 = getCentroid(templatesBySubject.get(c1));
            for (int c2 = c1 + 1; c2 < numberOfSubjects; c2++) {
                Template centroid2 = getCentroid(templatesBySubject.get(c2));
                double dist = getDistance(centroid1, centroid2);
                if (min > dist) {
                    min = dist;
                }
            }
        }
        double max = 0;
        for (int c = 0; c < numberOfSubjects; c++) {
            double dist = getAverageDistanceToCentroid(templatesBySubject.get(c));
            if (max < dist) {
                max = dist;
            }
        }
        return min / max;
    }

    public double getSilhouetteCoefficient(List<Template> templatesEvaluation) { // high is good
        List<List<Template>> templatesBySubject = new ClassifiableList(templatesEvaluation).splitBySubject();
        double sum = 0;
        for (List<Template> templatesOfThisSubject : templatesBySubject) {
            String thisSubject = templatesOfThisSubject.get(0).getSubject();
            for (Template template : templatesOfThisSubject) {
                double avgA = getAverageDistanceToTemplate(templatesOfThisSubject, template);
                double minAvgB = Double.MAX_VALUE;
                for (List<Template> templatesOfOtherSubject : templatesBySubject) {
                    String otherSubject = templatesOfOtherSubject.get(0).getSubject();
                    if (!thisSubject.equals(otherSubject)) {
                        double avgB = getAverageDistanceToTemplate(templatesOfOtherSubject, template);
                        if (minAvgB > avgB) {
                            minAvgB = avgB;
                        }
                    }
                }
                sum += (minAvgB - avgA) / Math.max(avgA, minAvgB);
            }
        }
        return sum / templatesEvaluation.size();
    }

    public double getFishersDiscriminantRatio(List<Template> templatesEvaluation) { // high is good
        List<List<Template>> templatesBySubject = new ClassifiableList(templatesEvaluation).splitBySubject();
        return getBetweenClassScatter(templatesBySubject) / getWithinClassScatter(templatesBySubject);
    }

    public double getPurity(List<List<Template>> clusters) { // high is good
        int numberOfTemplatesOfDominantSubjects = 0;
        int numberOfTemplatesOfAllSubjects = 0;
        for (List<Template> cluster : clusters) {
            List<List<Template>> templatesBySubject = new ClassifiableList(cluster).splitBySubject();
            int numberOfTemplatesOfDominantSubject = 0;
            for (List<Template> templatesOfSubject : templatesBySubject) {
                int numberOfTemplatesOfSubject = templatesOfSubject.size();
                if (numberOfTemplatesOfDominantSubject < numberOfTemplatesOfSubject) {
                    numberOfTemplatesOfDominantSubject = numberOfTemplatesOfSubject;
                }
            }
            numberOfTemplatesOfDominantSubjects += numberOfTemplatesOfDominantSubject;
            numberOfTemplatesOfAllSubjects += cluster.size();
        }
        return (double) numberOfTemplatesOfDominantSubjects / numberOfTemplatesOfAllSubjects;
    }

    public double getRandIndex(int tp, int fp, int fn, int tn) { // high is good
        return (double) (tp + tn) / (tp + fp + fn + tn);
    }

    public double getFmeasure(int tp, int fp, int fn, int tn) { // high is good
        double p = (double) tp / (tp + fp);
        double r = (double) tp / (tp + fn);
        return 2 * p * r / (p + r);
    }

    public double getJaccardIndex(int tp, int fp, int fn, int tn) { // high is good
        return (double) tp / (tp + fp + fn);
    }

    public double getFowlkesMallowsIndex(int tp, int fp, int fn, int tn) { // high is good
        return Math.sqrt(((double) tp / (tp + fp)) * ((double) tp / (tp + fn)));
    }

    private double getBetweenClassScatter(List<List<Template>> templatesBySubject) {
        List<Template> centroids = new ArrayList();
        for (List<Template> templatesOfSubject : templatesBySubject) {
            centroids.add(getCentroid(templatesOfSubject));
        }
        return getAverageDistanceToCentroid(centroids);
    }

    private double getWithinClassScatter(List<List<Template>> templatesBySubject) {
        double sum = 0;
        for (List<Template> templatesOfSubject : templatesBySubject) {
            sum += getAverageDistanceToCentroid(templatesOfSubject);
        }
        return sum / templatesBySubject.size();
    }

    private double getDistance(Template template1, Template template2) {
        return distanceMatrix.getDistance(template1, template2);
    }

    private double getAverageDistanceToCentroid(List<Template> templates) {
        return getAverageDistanceToTemplate(templates, getCentroid(templates));
    }

    private double getAverageDistanceToTemplate(List<Template> templates, Template template) {
        double sum = 0;
        for (Template otherTemplate : templates) {
            sum += getDistance(template, otherTemplate);
        }
        return sum / templates.size();
    }

    private Template getCentroid(List<Template> templates) {
        Template centroid = new Template(null, null);
        if (templates.isEmpty()) {
            System.out.println("empty class");
        } else {
            double minSum = Double.MAX_VALUE;
            for (Template centroidCandidate : templates) {
                double sum = 0;
                for (Template template : templates) {
                    sum += getDistance(template, centroidCandidate);
                }
                if (minSum > sum) {
                    minSum = sum;
                    centroid = centroidCandidate;
                }
            }
        }
        return centroid;
    }

    public int[] getContingency(List<List<Template>> clusters) {
        List<String> subjectsList = new ArrayList();
        List<Integer> clustersList = new ArrayList();
        for (int c = 0; c < clusters.size(); c++) {
            for (Template template : clusters.get(c)) {
                subjectsList.add(template.getSubject());
                clustersList.add(c);
            }
        }
        int numberOfTemplates = subjectsList.size();
        int tp = 0;
        int fp = 0;
        int fn = 0;
        int tn = 0;
        for (int i = 0; i < numberOfTemplates - 1; i++) {
            String subjectI = subjectsList.get(i);
            int clusterI = clustersList.get(i);
            for (int j = i + 1; j < numberOfTemplates; j++) {
                String subjectJ = subjectsList.get(j);
                int clusterJ = clustersList.get(j);
                boolean sameSubject = subjectI.equals(subjectJ);
                boolean sameCluster = clusterI == clusterJ;
                if (sameSubject && sameCluster) {
                    tp++;
                }
                if (!sameSubject && sameCluster) {
                    fp++;
                }
                if (sameSubject && !sameCluster) {
                    fn++;
                }
                if (!sameSubject && !sameCluster) {
                    tn++;
                }
            }
        }
        return new int[]{tp, fp, fn, tn};
    }

    public double getTemplateDimensionality(List<Template> templatesEvaluation) {
        int sum = 0;
        for (Template template : templatesEvaluation) {
            sum += template.getDimension() * template.getLength();
        }
        return (double) sum / templatesEvaluation.size();
    }
}
