package executor;

import algorithms.Classifier;
import algorithms.Classifier1NNDTWKrzeszowski;
import algorithms.Classifier1NNDTWL1;
import algorithms.Classifier1NNDTWL2;
import algorithms.Classifier1NNKumar;
import algorithms.Classifier1NNL1L1;
import algorithms.ClassifierRandom;
import algorithms.ClassifierTransform1NNMMCMahalanobis;
import algorithms.ClassifierTransform1NNPCALDAMahalanobis;
import algorithms.MethodEvaluatorClustering;
import algorithms.MethodEvaluatorClassification;
import algorithms.MotionLoaderBoneRotations;
import algorithms.MotionLoaderJointCoordinates;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.Scanner;
import objects.PoseJointCoordinates;
import objects.MotionJointCoordinates;
import objects.Feature;
import objects.Motion;
import methods.Method;
import methods.MethodAhmedM;
import methods.MethodAliS;
import methods.MethodAnderssonVO;
import methods.MethodBallA;
import methods.MethodDikovskiB;
import methods.MethodAhmedF;
import methods.MethodJiangS;
import methods.MethodKrzeszowskiT;
import methods.MethodNareshKumarMS;
import methods.MethodKwolekB;
import methods.MethodPreisJ;
import methods.MethodSedmidubskyJ;
import methods.MethodSinhaA;
import methods.MethodTransform;
import methods.MethodRandom;
import methods.MethodRaw;
import objects.ClassifiableList;
import objects.Parameter;
import objects.Template;

public class Executor {

    /*
     Select distance threshold
     Each value refers to a particular sub-database
     */
    private static final double distanceThreshold = 495.3; // 56.3(2,35), 59.4(4,67), 63.3(8,130), 73.7(16,302), 173.3(32,2047), 302.0(54,3843), 495.3(64,5923)

    /*
     Select actions
     */
    private static final boolean extractDatabase = false;
    private static final boolean learnClassifiers = false;
    private static final boolean performClassifications = false;
    private static final boolean evaluateMethods = true;

    /*
     Select evaluation set-ups at evaluateMethods()
     */
    private static final boolean evaluateClassification = false;
    private static final boolean evaluateClustering = true;

    /*
     Algorithms
     */
    private static final MotionLoaderBoneRotations motionLoaderBoneRotations = new MotionLoaderBoneRotations();
    private static final MotionLoaderJointCoordinates motionLoaderJointCoordinates = new MotionLoaderJointCoordinates();
    private static final Classifier1NNDTWL1 classifier1NNDTWL1 = new Classifier1NNDTWL1();
    private static final Classifier1NNDTWL2 classifier1NNDTWL2 = new Classifier1NNDTWL2();
    private static final Classifier1NNL1L1 classifier1NNL1L1 = new Classifier1NNL1L1();
    private static final ClassifierRandom classifierRandom = new ClassifierRandom();
    private static final Classifier1NNDTWKrzeszowski classifier1NNDTWKrzeszowski = new Classifier1NNDTWKrzeszowski();
    private static final Classifier1NNKumar classifier1NNKumar = new Classifier1NNKumar();
    private static final ClassifierTransform1NNMMCMahalanobis classifierTransform1NNMMCMahalanobis = new ClassifierTransform1NNMMCMahalanobis();
    private static final ClassifierTransform1NNPCALDAMahalanobis classifierTransform1NNPCALDAMahalanobis = new ClassifierTransform1NNPCALDAMahalanobis();

    /*
     All implemented MoCap-based gait recognition methods
     */
    private static Method[] methods = new Method[]{
        new MethodAhmedF(classifier1NNDTWL1),//0
        new MethodAhmedM(classifier1NNL1L1),//1---
        new MethodAliS(classifier1NNL1L1),//2---
        new MethodAnderssonVO(classifier1NNL1L1),//3---
        new MethodBallA(classifier1NNL1L1),//4---
        new MethodDikovskiB(classifier1NNL1L1),//5---
        new MethodJiangS(classifier1NNDTWL1),//6
        new MethodKrzeszowskiT(classifier1NNDTWKrzeszowski),//7
        new MethodKwolekB(classifier1NNL1L1),//8---
        new MethodNareshKumarMS(classifier1NNKumar),//9
        new MethodPreisJ(classifier1NNL1L1),//10---
        new MethodSedmidubskyJ(classifier1NNDTWL1),//11
        new MethodSinhaA(classifier1NNL1L1),//12---
        new MethodTransform(motionLoaderBoneRotations, classifierTransform1NNMMCMahalanobis),//13
        new MethodTransform(motionLoaderJointCoordinates, classifierTransform1NNMMCMahalanobis),//14---
        new MethodTransform(motionLoaderBoneRotations, classifierTransform1NNPCALDAMahalanobis),//15
        new MethodTransform(motionLoaderJointCoordinates, classifierTransform1NNPCALDAMahalanobis),//16---
        new MethodRandom(classifierRandom),//17
        new MethodRaw(motionLoaderBoneRotations, classifier1NNDTWL2),//18
        new MethodRaw(motionLoaderJointCoordinates, classifier1NNDTWL2),//19
    };

    /*
     Main method
     Executes the actions selected above
     */
    public static void main(String[] args) throws IOException, ClassNotFoundException {
        if (extractDatabase) {
            extractDatabase();
        }
        if (learnClassifiers) {
            learnClassifiers();
        }
        if (performClassifications) {
            performClassifications();
        }
        if (evaluateMethods) {
            evaluateMethods(args);
        }
    }

    /*
     Normalize the whole original CMU database in amcOriginal/ directory with respect to person’s position and walk direction and saves it to amc/ directory
     Select gait cycles as sub-motions with similarity under selected distanceThreshold to prototypicalFileAMC and save them to separate AMC files {subject}_{id}[{from}-{to}].amc in amc{distanceThreshold}/ directory
     Only subjects of at least 10 gait cycles are kept.
     */
    public static void extractDatabase() throws IOException {
        new File("amc/").mkdirs();
        for (File fileAMC : new File("amcOriginal/").listFiles()) {
            BufferedReader brAMC = new BufferedReader(new FileReader(fileAMC));
            BufferedWriter wrAMC = new BufferedWriter(new FileWriter("amc/" + fileAMC.getName()));
            String lineAMC = brAMC.readLine();
            while (lineAMC != null) {
                if (lineAMC.split(" ")[0].equals("root")) {
                    wrAMC.write("root 0 0 0 0 0 0\r\n");
                } else {
                    wrAMC.write(lineAMC + "\r\n");
                }
                lineAMC = brAMC.readLine();
            }
            wrAMC.flush();
            wrAMC.close();
        }
        MotionJointCoordinates motionProto = new MotionLoaderJointCoordinates().loadMotion(Constants.prototypicalFileAMC);
        Feature featureProto = motionProto.extractFeetDistanceSideSensitiveFeature();
        int lengthProto = featureProto.getLength();
        double sigma = 2;
        int minLength = (int) Math.round(lengthProto / sigma);
        int maxLength = (int) Math.round(lengthProto * sigma);
        List<Motion> gaitCycles = new ArrayList();
        for (File fileAMC : new File("amc/").listFiles()) {
            System.out.print("Processing " + fileAMC.getName() + "\r\n");
            MotionJointCoordinates motion = new MotionLoaderJointCoordinates().loadMotion(fileAMC);
            String identifier = motion.getIdentifier();
            Feature feature = motion.extractFeetDistanceSideSensitiveFeature();
            int length = feature.getLength();
            boolean selected = true;
            while (selected) {
                selected = false;
                double minDistance = Double.MAX_VALUE;
                int minFrom = 0;
                int minTo = 0;
                for (int from = 0; from < length; from++) {
                    for (int to = from + minLength; to < Math.min(from + maxLength, length); to++) {
                        Feature subFeature = feature.getSubFeature(from, to);
                        subFeature.adjust(lengthProto);
                        double[] subFeatureParameterValues = subFeature.getParameterValues();
                        double[] featureProtoParameterValues = featureProto.getParameterValues();
                        double distance = 0;
                        for (int l = 0; l < lengthProto; l++) {
                            distance += Math.abs(subFeatureParameterValues[l] - featureProtoParameterValues[l]);
                        }
                        if (minDistance > distance) {
                            minDistance = distance;
                            minFrom = from;
                            minTo = to;
                        }
                    }
                }
                if (minDistance < distanceThreshold) {
                    List<PoseJointCoordinates> subPoses = new ArrayList();
                    for (int l = minFrom; l < minTo && l < motion.getLength(); l++) {
                        subPoses.add((PoseJointCoordinates) motion.getPose(l));
                    }
                    gaitCycles.add(new MotionJointCoordinates(identifier + "_" + (minFrom + 1) + "-" + (minTo + 1), subPoses));
                    selected = true;
                    for (int l = minFrom; l < minTo; l++) {
                        feature.setParameter(l, new Parameter("Erased", new Random().nextDouble() * Double.MAX_VALUE));
                    }
                }
            }
        }
        List<List<Motion>> gaitCyclesBySubject = new ClassifiableList(gaitCycles).splitBySubject();
        new File("amc" + distanceThreshold + "/").mkdirs();
        for (List<Motion> gaitCyclesOfSubject : gaitCyclesBySubject) {
            if (gaitCyclesOfSubject.size() >= 10) {
                for (Motion gaitCycle : gaitCyclesOfSubject) {
                    String identifier = gaitCycle.getIdentifier();
                    BufferedReader brAMC = new BufferedReader(new FileReader("amc/" + identifier + ".amc"));
                    BufferedWriter wrAMC = new BufferedWriter(new FileWriter("amc" + distanceThreshold + "/" + identifier + ".amc"));
                    String subSequence = identifier.split("_")[2];
                    String from = subSequence.split("-")[0];
                    String to = subSequence.split("-")[1];
                    String lineAMC = brAMC.readLine();
                    while (lineAMC != null && !lineAMC.equals("1")) {
                        wrAMC.write(lineAMC + "\r\n");
                        lineAMC = brAMC.readLine();
                    }
                    while (lineAMC != null && !lineAMC.equals(String.valueOf(from))) {
                        lineAMC = brAMC.readLine();
                    }
                    int f = 0;
                    while (lineAMC != null && !lineAMC.equals(to)) {
                        if (lineAMC.split(" ").length == 1) {
                            wrAMC.write(++f + "\r\n");
                        } else {
                            wrAMC.write(lineAMC + "\r\n");
                        }
                        lineAMC = brAMC.readLine();
                    }
                    wrAMC.flush();
                    wrAMC.close();
                    brAMC.close();
                }
            }
        }
    }

    /*
     Extracts classifiers for all methods learned on the sub-database determined by distanceThreshold
     Classifiers are saved as separate files {methodName}-{distanceThreshold}.classifier in classifiers/ directory
     */
    private static void learnClassifiers() throws IOException {
        new File("classifiers/").mkdirs();
        for (Method method : methods) {
            method.learnClassifier(method.loadMotions(new File("amc" + distanceThreshold + "/")));
            method.saveClassifier(new File("classifiers/" + method.getName() + "-" + distanceThreshold + ".classifier"));
        }
    }

    /*
     Performs classifications of customProbeFileAMC over customGalleryDirectory for all methods
     Instead of learning classifiers methods load their extracted classifiers {distanceThreshold}-{methodName}.classifier in classifiers/ directory that are learned on the sub-database determined by distanceThreshold
     */
    private static void performClassifications() throws IOException, ClassNotFoundException {
        ObjectInputStream oos = new ObjectInputStream(new FileInputStream(Constants.customClassifier));
        Classifier classifier = (Classifier) oos.readObject();
        oos.close();
        for (Method method : methods) {
            classifier.getDecision().importGallery(method.extractTemplates(method.loadMotions(Constants.customGalleryDirectory)));
            Template templateQuery = method.extractTemplate(method.loadMotion(Constants.customProbeFileAMC));
            System.out.print(method.getName() + ": query subject classified as " + classifier.classify(templateQuery) + "\r\n");
        }
    }

    /*
     Performs evaluations for all methods
     */
    private static void evaluateMethods(String[] args) throws IOException {
        long beginning = System.currentTimeMillis();

        if (evaluateClassification) {
            //default
            int attempts = 1; // number of evaluation attempts
            int nFolds = 10; // number of cross-validation folds
            methods = new Method[]{methods[Integer.parseInt(args[0])]};//Integer.parseInt(args[0])
            MethodEvaluatorClassification methodEvaluator = new MethodEvaluatorClassification(methods, new File("amc" + distanceThreshold + "/"), nFolds);
            System.out.print(distanceThreshold + ",,CCR\r\n");
            for (int i = Integer.parseInt(args[1]); i <= Integer.parseInt(args[2]); i++) {
                int nClasses = i;
                System.out.print("[" + nClasses + "],");
                methodEvaluator.setSubjects(nClasses);
                methodEvaluator.initialize();
                for (int attempt = 0; attempt < attempts; attempt++) {
                    methodEvaluator.evaluate();
                }
                methodEvaluator.print();
            }
//            // Svenda
//            method.evaluateClassification(motions, numberOfClasses, 10, 1, 30);
//            // S+SSPR Figure 3
//            for (int i = 2; i <= numberOfClasses / 2; i++) {
//                nClasses = i;
//                nFoldsLearning = 3;
//                attempts = 1;
//                System.out.print("(" + nClasses + "," + nClasses +")\r\n");
//                beginning = System.currentTimeMillis();
//                method.evaluateClassification(motions, nClasses, nFoldsLearning, attempts);
//            }
//            // ACM-TOMM
//            System.out.print("CCR,EER,AUC,MAP,DCT,TD\r\n");
//            new MethodEvaluator(method).evaluateClassification(motions, numberOfClasses, 10, 1);
        }
        if (evaluateClustering) {
//            // default
//            nClassesLearning = 2; // number of learning classes
//            nClassesEvaluation = numberOfClasses - 2;
//            attempts = 3;
//            method.evaluateClustering(motions, nClassesLearning, nClassesEvaluation, attempts);
//            System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//            // Svenda
//            System.out.print("DBI,DI,SC,FDR,CCR,EER,AUC,MAP,DCT,TD\r\n");
//            method.evaluateClassification(motions, 5, 2, 1);
//            // S+SSPR Figure 3
//            for (int i = 2; i <= numberOfClasses / 2; i++) {
//                nClassesLearning = i;
//                nClassesEvaluation = i;
//                attempts = 3;
//                System.out.print("(" + nClassesLearning + ";" + nClassesEvaluation +")\r\n");
//                method.evaluateClustering(motions, nClassesLearning, nClassesEvaluation, attempts);
//                System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//            }
//            // S+SSPR Figure 4a
//            for (int i = 2; i <= numberOfClasses / 2; i++) {
//                nClassesLearning = i;
//                nClassesEvaluation = numberOfClasses / 2;
//                attempts = 3;
//                System.out.print("(" + nClassesLearning + ";" + nClassesEvaluation +")\r\n");
//                method.evaluateClustering(motions, nClassesLearning, nClassesEvaluation, attempts);
//                System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//            }
//            // S+SSPR Figure 4b
//            for (int i = 2; i <= numberOfClasses - 2; i++) {
//                nClassesLearning = i;
//                nClassesEvaluation = numberOfClasses - i;
//                attempts = 3;
//                System.out.print("(" + nClassesLearning + ";" + nClassesEvaluation + ")\r\n");
//                method.evaluateClustering(motions, nClassesLearning, nClassesEvaluation, attempts);
//                System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//            }
            // IJCB
            int attempts = 1; // number of evaluation attempts
            int fineness = 30; // number of thresholds for FAR/FRR TAR/FAR RCL/PCN
            System.out.print(distanceThreshold + "\r\n");
            System.out.print("method,DCT,TD,DBI,DI,SC,FDR,EER,ROC,PR,P-Kmeans,P-AHC,RI-Kmeans,RI-AHC,F-Kmeans,F-AHC,JI-Kmeans,JI-AHC,FMI-Kmeans,FMI-AHC\r\n");
            methods = new Method[]{methods[Integer.parseInt(args[0])]};//Integer.parseInt(args[0])
            MethodEvaluatorClustering methodEvaluator = new MethodEvaluatorClustering(methods, new File("amc" + distanceThreshold), fineness);
//            // random
//            int nClassesLearning = 2;//Integer.parseInt(args[0]);
//            int nClassesEvaluation = new ClassifiableList(methodEvaluator.getClassifiables()).splitBySubject().size() - nClassesLearning;//
//            methodEvaluator.setSubjects(nClassesLearning, nClassesEvaluation);//
            // 9-55
            List<String> subjectsLearning = transformToSubjects("[127; 13; 133; 78; 107; 19; 08; 63; 91]");//
            List<String> subjectsEvaluation = transformToSubjects("[106; 85; 88; 81; 69; 54; 15; 38; 35; 104; 135; 136; 94; 17; 138; 114; 126; 111; 132; 77; 93; 108; 137; 74; 143; 16; 22; 01; 05; 14; 90; 03; 61; 89; 39; 113; 36; 09; 131; 49; 86; 142; 40; 41; 82; 60; 06; 120; 56; 55; 70; 83; 18; 12; 07]");//
            methodEvaluator.setSubjects(subjectsLearning, subjectsEvaluation);

            methodEvaluator.initialize();
            for (int attempt = 0; attempt < attempts; attempt++) {
                methodEvaluator.evaluate();
            }
            methodEvaluator.print();
        }
        System.out.print("time=" + printTime(System.currentTimeMillis() - beginning));
    }

    private static List<String> transformToSubjects(String string) {
        string = string.replace("[", "");
        string = string.replace("]", "");
        Scanner scanner = new Scanner(string).useDelimiter(";\\s*");
        List<String> subjects = new ArrayList<String>();
        while (scanner.hasNext()) {
            String subject = scanner.next();
            subjects.add(subject);
        }
        return subjects;
    }

    private static String printTime(double time) {
        String unit = "ms";
        if (time >= 1000) {
            time /= 1000;
            unit = "s";
            if (time >= 60) {
                time /= 60;
                unit = "m";
                if (time >= 60) {
                    time /= 60;
                    unit = "h";
                    if (time >= 24) {
                        time /= 24;
                        unit = "d";
                    }
                }
            }
        }
        return String.valueOf((double) Math.round(time * 10) / 10) + unit;
    }
}
